/*
 * Rehacking it An example of the usage for standard ethernet is: 
 *
 * cdlpi -c 5 -d 2:60:8c:2e:aa:3c -f 1 -n /dev/dlpi/en -s 0x660 
 *
 */


/*
 * cdlpi - dlpi client 
 */

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>

#include <sys/poll.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/dlpi.h>
#include <sys/dlpistats.h>
#include <netinet/if_ether.h>
#include <sys/cdli.h>

char            data[4096];	/* data portion of response */
char            buf[4096];	/* proto portion of response */

char           *conv();
int             fd;

/* static so init to all zeros */

static dl_attach_req_t attach;
static dl_bind_req_t mybind;
static dl_info_req_t info;
static dl_unitdata_req_t *udreq;
static dl_unitdata_ind_t *udind;

int pkt_format;
        char            src_address[64];
        ulong           src_addr_sap_len;
char *address;

setup_dlpi()
{
	int             c, i, j, num_pack = 0;
	int             flags, icnt, count = 1, total = 0;
	struct strbuf   ctl, dat;
	long            saplen;
	int             sflag = 0, nflag = 0, dflag = 0, fflag = 0;
	struct strioctl strioctl;

	dl_ok_ack_t    *okack;
	dl_bind_ack_t  *bindack;
	dl_info_ack_t  *infoack;

	char           *node, *p;

	node = "/dev/dlpi/en";
	pkt_format = NS_INCLUDE_MAC; /* default */

	icnt = count;

	if ((fd = open(node, O_RDWR)) < 0)
		die(node);

	/*
	 * attach to interface 
	 */

	attach.dl_primitive = DL_ATTACH_REQ;
	attach.dl_ppa = 0;

	(void) request(fd, &attach, sizeof attach, buf, sizeof buf, "attach");

	okack = (dl_ok_ack_t *) buf;
	if (okack->dl_primitive != DL_OK_ACK)
		errorack(buf, "cannot attach");

	printf("dlpi.o: attached\n");
   /*
    * Set up the packet format registration.  This must be specified after
    * the attach and before the bind.  NS_PROTO is the default packet type,
    * so we do not need to do anything if we are using the default.
    */

   if (pkt_format != NS_PROTO) {
      strioctl.ic_dp = (char *)&pkt_format;
      strioctl.ic_cmd = DL_PKT_FORMAT;
      strioctl.ic_timout = 0;
      strioctl.ic_len = sizeof(long);

      if (ioctl(fd, I_STR, &strioctl) < 0) {
         perror("ioctl I_STR failed");
         exit(1);
      }
   }


	/*
	 * bind to sap 
	 */

	mybind.dl_primitive = DL_BIND_REQ;
	mybind.dl_sap = 0xaaaa;
	mybind.dl_max_conind = 0;
	mybind.dl_service_mode = DL_CLDLS;
	mybind.dl_conn_mgmt = 0;
	mybind.dl_xidtest_flg = 0;

	(void) request(fd, &mybind, sizeof mybind, buf, sizeof buf, "bind");

	bindack = (dl_bind_ack_t *) buf;
	if (bindack->dl_primitive != DL_BIND_ACK)
		errorack(buf, "cannot bind");

	/*
	 * Save the source address in the event the packet format has been
	 * set to NS_INCLUDE_MAC.  We will generically determine the address
	 * length below using the sap length returned in the dl_info_ack_t. 
	 */

	src_addr_sap_len = bindack->dl_addr_length;
	bcopy(buf + bindack->dl_addr_offset, src_address, bindack->dl_addr_length);

	printf("dpli.o: bound to:\n");
	printf("  sap       = %x\n", bindack->dl_sap);
	printf("  address   = ");
	p = buf + bindack->dl_addr_offset;
	for (i = bindack->dl_addr_length; i--;)
		printf("%02x:", *p++);
	printf("\n");

	/*
	 * get info 
	 */

	info.dl_primitive = DL_INFO_REQ;

	(void) request(fd, &info, sizeof info, buf, sizeof buf, "info");

	infoack = (dl_info_ack_t *) buf;
	if (infoack->dl_primitive != DL_INFO_ACK)
		errorack(buf, "cannot get info");

	/*
	 * Save the sap length from the DL_INFO_ACK so that we can
	 * generically determine the source address length.  This is used if
	 * a packet format other than NS_PROTO is specified. 
	 */

	saplen = infoack->dl_sap_length;


	printf("dlpi.o: provider info:\n");
	printf("  max sdu       = %d\n", infoack->dl_max_sdu);
	printf("  min sdu       = %d\n", infoack->dl_min_sdu);
	printf("  mac type      = %d\n", infoack->dl_mac_type);
	printf("  addr length   = %d\n", infoack->dl_addr_length);
	printf("  current state = %d\n", infoack->dl_current_state);
	printf("  service mode  = %d\n", infoack->dl_service_mode);
	printf("\n");
}

/*
 * client request 
 */
void 
dlpi_send(char *ether_adr, char *mydata, long size)
{
	struct strbuf   ctl, dat;
	struct pollfd   fds;
	int             nfds, i;
	char *p;

        address = conv(ether_adr);

	udreq = (dl_unitdata_req_t *) buf;
	udreq->dl_primitive = DL_UNITDATA_REQ;
	udreq->dl_dest_addr_offset = sizeof(dl_unitdata_req_t);
	udreq->dl_priority.dl_min = udreq->dl_priority.dl_max = 0;

	/*
	 * Build the header as specified by the packet format. NOTE THAT
	 * THESE EXAMPLES ARE FOR STANDARD ETHERNET! 
	 */

	switch (pkt_format) {
	case NS_PROTO:
		udreq->dl_dest_addr_length = 6;
		memcpy(buf + udreq->dl_dest_addr_offset, address,
		       udreq->dl_dest_addr_length);

		/*
		 * print out the header 
		 */

		p = buf + udreq->dl_dest_addr_offset;
/*		for (i = udreq->dl_dest_addr_length; i--;)
			printf("%02x:", *p++);
		printf("\n"); */

		ctl.len = sizeof(dl_unitdata_req_t) + 6;
		/* Ip Type Diagram */
		data[0]=8;
		data[1]=0;
		(void)bcopy( mydata, data+2, size );
		dat.len = size + 2;
		break;

	case NS_PROTO_SNAP:
		die("NS_PROTO_SNAP is not applicable to standard ethernet");

	case NS_INCLUDE_LLC:
		die("NS_INCLUDE_LLC is not applicable to standard ethernet");

	case NS_INCLUDE_MAC:
		/*
		 * Build the MAC header for standard ethernet.  You need to
		 * know what interface type you are using so that you can
		 * build the proper MAC header. 
		 */

		/* place the destination address in the data portion */
		p = data;
		bcopy(address, p, 6);
		/*
		 * place the source address and sap in the data portion 
		 */
		p += 6;
/*		bcopy(src_address, p, src_addr_sap_len);
		7* copy in the data *7
		p += src_addr_sap_len;
		*p++ = 3; 7* Control *7
		*p++ = 0; 7* Org code *7
		*p++ = 0;
		*p++ = 0;
		*((short *)p) = 0x0800; p += 2;
		(void)bcopy( mydata, p, size );
		dat.len = 2+3+1+6+sizeof(src_addr_sap_len)+size; */
		bcopy(src_address, p, 6);
		p += 6;
		*((short *)p) = 0x0800; p += 2;
		(void)bcopy( mydata, p, size );
                dat.len = 2+6+6+size;


		/*
		 * print out the MAC header 
		 */

/*		p = data;
		for (i = 6 + src_addr_sap_len; i--;)
			printf("%02x:", *p++);
		printf("\n"); */

		ctl.len = sizeof(dl_unitdata_req_t);
		break;
	}

	ctl.buf = buf;
	dat.buf = data;
	dat.maxlen = sizeof data;

	if (putmsg(fd, &ctl, &dat, 0))
		die("putmsg");
}

errorack(p, msg)
dl_error_ack_t * p;
	char           *msg;
{
	extern int      errno;

	if (p->dl_errno != DL_SYSERR)
		fprintf(stderr, "%s: dl_errno %d\n", msg, p->dl_errno);
	else {
		errno = p->dl_unix_errno;
		perror(msg);
	}
	exit(1);
}

/*
 * request - make a DLPI request 
 *
 * returns length of user data stores proto data in res, user data in data[] 
 */

int
request(fd, req, req_len, res, res_len, foo)
	int             fd;
	caddr_t         req;
	int             req_len;
	caddr_t         res;
	int             res_len;
	char           *foo;
{
	struct strbuf   ctl, dat;
	int             flags = 0;

	ctl.buf = req;
	ctl.len = req_len;
	dat.len = 0;
	dat.buf = data;

	if (putmsg(fd, &ctl, &dat, 0))
		die("putmsg - %s", foo);

	ctl.buf = res;
	ctl.maxlen = res_len;
	dat.len = 0;
	dat.buf = data;
	dat.maxlen = sizeof data;

	if (getmsg(fd, &ctl, &dat, &flags) != 0)
		die("getmsg - %s", foo);

	return dat.len;
}

char           *
conv(cp)
	char           *cp;
{
	static char     buf[100];
	char           *addr;
	char           *p;

	addr = buf;
	for (p = strtok(cp, ":"); p; p = strtok(0, ":"))
		*addr++ = strtoul(p, 0, 16);
	return buf;
}

die(s, a, b, c)
	char           *s;
{
	char            buf[256];
	sprintf(buf, s, a, b, c);
	perror(buf);
	exit(0);
}
